"""Configuration management functions."""

from collections import defaultdict
import dataclasses
from pathlib import Path
import typing as t

from prometheus_aioexporter import MetricConfig
from pydantic import ValidationError
import structlog
import yaml

from . import db, schema
from .metrics import BUILTIN_METRICS, get_builtin_metric_configs
from .yaml import load_yaml

# Label used to tag metrics by database
DATABASE_LABEL = "database"


class ConfigError(Exception):
    """Configuration is invalid.

    Optionally, a list of dicts can be provided with error details.
    """

    def __init__(self, message: str, details: t.Sequence[dict[str, str]] = ()):
        super().__init__(message)
        self.details = details


@dataclasses.dataclass(frozen=True)
class Config:
    """Top-level configuration."""

    databases: dict[str, schema.Database]
    metrics: dict[str, MetricConfig]
    queries: dict[str, db.Query]


def load_config(
    paths: list[Path],
    logger: structlog.stdlib.BoundLogger | None = None,
) -> Config:
    """Load YAML configuration files."""
    if logger is None:
        logger = structlog.get_logger()

    try:
        data = _load_config(paths)
    except yaml.scanner.ScannerError as e:
        raise ConfigError(str(e))
    try:
        configuration = schema.ExporterConfig(**data)
    except ValidationError as ve:
        errors = ve.errors()
        details = [
            {
                "location": (
                    ".".join(str(item) for item in e["loc"])
                    if e["loc"]
                    else ""
                ),
                "error": e["msg"],
            }
            for e in errors
        ]
        raise ConfigError(
            f"{len(errors)} validation error{'s' if len(errors) > 1 else ''}",
            details=details,
        )
    database_labels = _validate_databases(configuration.databases)
    extra_labels = frozenset([DATABASE_LABEL]) | database_labels
    builtin_metrics_config = (
        {
            name: metric.config()
            for name, metric in configuration.builtin_metrics.as_dict().items()
        }
        if configuration.builtin_metrics
        else {}
    )
    metrics = _get_metrics(
        configuration.metrics, builtin_metrics_config, extra_labels
    )
    queries = _get_queries(
        configuration.queries,
        frozenset(configuration.databases),
        metrics,
        extra_labels,
    )
    config = Config(configuration.databases, metrics, queries)
    _warn_if_unused(config, logger)
    return config


def _load_config(paths: list[Path]) -> dict[str, t.Any]:
    """Return the combined configuration from provided files."""
    config: dict[str, t.Any] = defaultdict(dict)
    for path in paths:
        conf = load_yaml(path)
        if not isinstance(conf, dict):
            raise ConfigError(f"File content is not a mapping: {path}")
        data = defaultdict(dict, conf)
        for key in (field.name for field in dataclasses.fields(Config)):
            entries = data.pop(key, None)
            if entries is not None:
                if overlap_entries := set(config[key]) & set(entries):
                    overlap_list = ", ".join(sorted(overlap_entries))
                    raise ConfigError(
                        f'Duplicated entries in the "{key}" section: {overlap_list}'
                    )
                config[key].update(entries)
        config.update(data)
    return config


def _validate_databases(dbs: dict[str, schema.Database]) -> frozenset[str]:
    """Validate database configuration and return a set of all database labels."""
    all_db_labels: set[frozenset[str]] = set()  # set of all labels sets
    try:
        for name, config in dbs.items():
            # raise an error if the configuration is not valid
            db.create_db_engine(config)
            all_db_labels.add(frozenset(config.labels))
    except Exception as e:
        raise ConfigError(str(e))

    db_labels: frozenset[str]
    if not all_db_labels:
        db_labels = frozenset()
    elif len(all_db_labels) > 1:
        raise ConfigError("Not all databases define the same labels")
    else:
        db_labels = all_db_labels.pop()

    return db_labels


def _get_metrics(
    metrics: dict[str, schema.Metric],
    builtin_metrics_config: dict[str, dict[str, t.Any]],
    extra_labels: frozenset[str],
) -> dict[str, MetricConfig]:
    """Return a dict mapping metric names to their configuration."""
    configs = get_builtin_metric_configs(extra_labels, builtin_metrics_config)
    for name, metric in metrics.items():
        _validate_metric_config(name, metric, extra_labels)
        configs[name] = MetricConfig(
            name,
            metric.description,
            metric.type,
            labels=list(set(metric.labels) | extra_labels),
            config=metric.config,
        )
    return configs


def _validate_metric_config(
    name: str, metric: schema.Metric, extra_labels: frozenset[str]
) -> None:
    """Validate a metric configuration stanza."""
    if name in BUILTIN_METRICS:
        raise ConfigError(f'Label name "{name} is reserved for builtin metric')
    labels = set(metric.labels)
    overlap_labels = labels & extra_labels
    if overlap_labels:
        overlap_list = ", ".join(sorted(overlap_labels))
        raise ConfigError(
            f'Labels for metric "{name}" overlap with reserved/database ones: {overlap_list}'
        )


def _get_queries(
    configs: dict[str, schema.Query],
    database_names: frozenset[str],
    metrics: dict[str, MetricConfig],
    extra_labels: frozenset[str],
) -> dict[str, db.Query]:
    """Return a list of Queries from config."""
    metric_names = frozenset(metrics)
    queries: dict[str, db.Query] = {}
    for name, config in configs.items():
        _validate_query_config(name, config, database_names, metric_names)
        query_metrics = _get_query_metrics(config, metrics, extra_labels)
        parameter_sets = t.cast(list[dict[str, t.Any]], config.parameters)
        try:
            queries[name] = db.Query(
                name=name,
                databases=config.databases,
                metrics=query_metrics,
                sql=config.sql.strip(),
                timeout=config.timeout,
                interval=config.interval,
                schedule=config.schedule,
                parameter_sets=parameter_sets,
            )
        except (db.InvalidQueryParameters, db.InvalidQuerySchedule) as e:
            raise ConfigError(str(e))
    return queries


def _get_query_metrics(
    query: schema.Query,
    metrics: dict[str, MetricConfig],
    extra_labels: frozenset[str],
) -> list[db.QueryMetric]:
    """Return QueryMetrics for a query."""

    def _metric_labels(labels: t.Iterable[str]) -> list[str]:
        return sorted(set(labels) - extra_labels)

    return [
        db.QueryMetric(name, _metric_labels(metrics[name].labels))
        for name in query.metrics
    ]


def _validate_query_config(
    name: str,
    query: schema.Query,
    database_names: frozenset[str],
    metric_names: frozenset[str],
) -> None:
    """Validate a query configuration stanza."""
    unknown_databases = set(query.databases) - database_names
    if unknown_databases:
        unknown_list = ", ".join(sorted(unknown_databases))
        raise ConfigError(
            f'Unknown databases for query "{name}": {unknown_list}'
        )
    unknown_metrics = set(query.metrics) - metric_names
    if unknown_metrics:
        unknown_list = ", ".join(sorted(unknown_metrics))
        raise ConfigError(
            f'Unknown metrics for query "{name}": {unknown_list}'
        )
    if query.parameters:
        params = t.cast(list[dict[str, t.Any]], query.parameters)
        keys = {frozenset(param.keys()) for param in params}
        if len(keys) > 1:
            raise ConfigError(
                f'Invalid parameters definition for query "{name}": '
                "parameters dictionaries must all have the same keys"
            )


def _warn_if_unused(
    config: Config, logger: structlog.stdlib.BoundLogger
) -> None:
    """Warn if there are unused databases or metrics defined."""
    used_dbs: set[str] = set()
    used_metrics: set[str] = set()
    for query in config.queries.values():
        used_dbs.update(query.databases)
        used_metrics.update(metric.name for metric in query.metrics)

    if unused_dbs := sorted(set(config.databases) - used_dbs):
        logger.warning(
            "unused config entries",
            section="databases",
            entries=unused_dbs,
        )
    if unused_metrics := sorted(
        set(config.metrics) - BUILTIN_METRICS - used_metrics
    ):
        logger.warning(
            "unused config entries",
            section="metrics",
            entries=unused_metrics,
        )



from enum import StrEnum
from functools import reduce
from itertools import product
import re
import typing as t

from pydantic import (
    AfterValidator,
    AliasGenerator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
)
from sqlalchemy import exc
from sqlalchemy.engine import URL, make_url


def _validate_unique_items(items: list[t.Any]) -> list[t.Any]:
    """Validate that there are no duplicate items in the list."""
    assert sorted(items) == sorted(set(items)), (
        "must not contain duplicate items"
    )
    return items


def _validate_sorted(items: list[t.Any]) -> list[t.Any]:
    """Validate that items in the list are sorted."""
    assert items == sorted(items), "items must be sorted"
    return items


_INTERVAL_RE = re.compile(r"^[0-9]+[smhd]?$")


def _validate_interval(interval: int | str) -> int:
    """Convert a time interval to seconds.

    Return None if no interval is specified.

    """
    multiplier = 1
    if isinstance(interval, str):
        assert _INTERVAL_RE.match(interval), "invalid interval definition"
        # convert to seconds
        multipliers = {"s": 1, "m": 60, "h": 3600, "d": 3600 * 24}
        suffix = interval[-1]
        if suffix in multipliers:
            interval = interval[:-1]
            multiplier = multipliers[suffix]

    value = int(interval) * multiplier
    assert value > 0, "must be a positive number"
    return value


Buckets = t.Annotated[
    list[float],
    Field(min_length=1),
    AfterValidator(_validate_unique_items),
    AfterValidator(_validate_sorted),
]
Label = t.Annotated[
    str,
    Field(pattern="^[a-zA-Z_][a-zA-Z0-9_]*$"),
]
TimeInterval = t.Annotated[
    int,
    BeforeValidator(_validate_interval),
]
Timeout = t.Annotated[
    float,
    Field(
        gt=0,
        multiple_of=0.1,
    ),
]
Port = t.Annotated[
    int,
    Field(
        ge=1,
        le=65535,
    ),
]


class MetricType(StrEnum):
    """Supported metric types."""

    COUNTER = "counter"
    ENUM = "enum"
    GAUGE = "gauge"
    HISTOGRAM = "histogram"
    SUMMARY = "summary"


class Model(BaseModel):
    """Base schema model."""

    model_config = ConfigDict(
        # support dashed field names
        alias_generator=AliasGenerator(alias=lambda s: s.replace("_", "-")),
        # don't allow extra fields
        extra="forbid",
    )


class DSNDetails(Model):
    """DSN configuration details."""

    dialect: str
    user: str | None = None
    password: str | None = None
    host: str | None = None
    port: Port | None = None
    database: str | None = None
    options: dict[str, str | list[str]] = Field(default_factory=dict)


def _validate_dsn(dsn: str | dict[str, str | list[str]]) -> str:
    """Build the DSN string from its definition."""
    url: URL
    try:
        match dsn:
            case str():
                url = make_url(dsn)
            case DSNDetails():
                url = URL.create(
                    dsn.dialect,
                    username=dsn.user,
                    password=dsn.password,
                    host=dsn.host,
                    port=dsn.port,
                    database=dsn.database,
                    query=dsn.options,
                )
    except exc.ArgumentError:
        raise ValueError("Invalid DSN format")
    return url.render_as_string(hide_password=False)


DSN = t.Annotated[str | DSNDetails, AfterValidator(_validate_dsn)]


class BuiltinMetric(Model):
    """Configuration for a builtin metric"""

    buckets: Buckets

    def config(self) -> dict[str, t.Any]:
        """The metric configuration."""

        return self.model_dump()


class BuiltinMetrics(BaseModel):
    """Configuration for builtin metrics"""

    model_config = ConfigDict(extra="forbid")

    query_latency: BuiltinMetric | None = None

    def as_dict(self) -> dict[str, BuiltinMetric]:
        """Return builtin metrics as a dictionary keyed by metric name."""
        d = {}
        for name in self.__class__.model_fields:
            if value := getattr(self, name):
                d[name] = value
        return d


class ConnectionPool(Model):
    """Database connection pool configuration."""

    size: int = Field(ge=1, le=100, default=1)
    max_overflow: int = Field(ge=0, le=100, default=0)


class Database(Model):
    """Database connection configuration."""

    dsn: DSN
    connection_pool: ConnectionPool = Field(default_factory=ConnectionPool)
    connect_sql: list[str] = Field(min_length=1, default_factory=list)
    labels: dict[Label, str] = Field(min_length=1, default_factory=dict)


class Metric(Model):
    """A metric definition."""

    type: MetricType
    description: str = ""
    labels: list[Label] = Field(min_length=1, default_factory=list)
    buckets: Buckets | None = None
    states: (
        t.Annotated[
            list[str],
            Field(min_length=1),
            AfterValidator(_validate_unique_items),
        ]
        | None
    ) = None
    expiration: TimeInterval | None = None
    increment: bool = False

    @model_validator(mode="after")
    def validate_all(self) -> t.Self:
        if self.states and not self.type == MetricType.ENUM:
            raise ValueError("states can only be set for enum metrics")
        if self.increment and not self.type == MetricType.COUNTER:
            raise ValueError("increment can only be set for counter metrics")
        return self

    @property
    def config(self) -> dict[str, t.Any]:
        """The metric configuration."""
        return self.model_dump(
            exclude={"type", "description", "labels"},
            exclude_none=True,
            exclude_defaults=True,
        )


def _validate_query_paramters(
    parameters: list[dict[str, t.Any]] | dict[str, list[dict[str, t.Any]]],
) -> list[dict[str, t.Any]]:
    """Return an sequence of set of paramters with their values."""
    if isinstance(parameters, list):
        return parameters

    # first, flatten dict like
    #
    # {
    #     'a': [{'arg1': 1, 'arg2': 1}],
    #     'b': [{'arg1': 1, 'arg2': 1}],
    # }
    #
    # into a sequence like
    #
    # (
    #     [{'a__arg1': 1, 'a__arg2': 2}],
    #     [{'b__arg1': 1, 'b__arg2': 2}],
    # )
    flattened_params = (
        [
            {f"{top_key}__{key}": value for key, value in arg_set.items()}
            for arg_set in arg_sets
        ]
        for top_key, arg_sets in parameters.items()
    )
    # return a list of merged dictionaries from each combination of the two
    # sets
    return list(
        reduce(lambda p1, p2: {**p1, **p2}, params)
        for params in product(*flattened_params)
    )


QueryParameters = t.Annotated[
    t.Annotated[list[dict[str, t.Any]], Field(min_length=1)]
    | t.Annotated[dict[str, list[dict[str, t.Any]]], Field(min_length=1)],
    AfterValidator(_validate_query_paramters),
]


class Query(Model):
    """A database query definition."""

    databases: t.Annotated[
        list[str], Field(min_length=1), AfterValidator(_validate_unique_items)
    ]
    metrics: t.Annotated[
        list[str], Field(min_length=1), AfterValidator(_validate_unique_items)
    ]
    sql: str
    interval: TimeInterval | None = None
    parameters: QueryParameters | None = None
    schedule: str | None = None
    timeout: Timeout | None = None


class ExporterConfig(Model):
    """Exporter configuration."""

    builtin_metrics: BuiltinMetrics | None = None
    databases: dict[str, Database]
    metrics: dict[Label, Metric]
    queries: dict[str, Query]



import typing as t

from prometheus_aioexporter import MetricConfig

# metric for counting database errors
DB_ERRORS_METRIC_NAME = "database_errors"
_DB_ERRORS_METRIC_CONFIG = MetricConfig(
    name=DB_ERRORS_METRIC_NAME,
    description="Number of database errors",
    type="counter",
    config={"increment": True},
)

# metric for counting performed queries
QUERIES_METRIC_NAME = "queries"
_QUERIES_METRIC_CONFIG = MetricConfig(
    name=QUERIES_METRIC_NAME,
    description="Number of database queries",
    type="counter",
    labels=("query", "status"),
    config={"increment": True},
)
# metric for tracking last query execution timestamp
QUERY_TIMESTAMP_METRIC_NAME = "query_timestamp"
_QUERY_TIMESTAMP_METRIC_CONFIG = MetricConfig(
    name=QUERY_TIMESTAMP_METRIC_NAME,
    description="Query last execution timestamp",
    type="gauge",
    labels=("query",),
)
# metric for counting queries execution latency
QUERY_LATENCY_METRIC_NAME = "query_latency"
_QUERY_LATENCY_METRIC_CONFIG = MetricConfig(
    name=QUERY_LATENCY_METRIC_NAME,
    description="Query execution latency",
    type="histogram",
    labels=("query",),
)
# metrics reporting the query interval
QUERY_INTERVAL_METRIC_NAME = "query_interval"
_QUERY_INTERVAL_METRIC_CONFIG = MetricConfig(
    name=QUERY_INTERVAL_METRIC_NAME,
    description="Query execution interval",
    type="gauge",
    labels=("query",),
)


BUILTIN_METRICS = frozenset(
    (
        DB_ERRORS_METRIC_NAME,
        QUERIES_METRIC_NAME,
        QUERY_INTERVAL_METRIC_NAME,
        QUERY_LATENCY_METRIC_NAME,
        QUERY_TIMESTAMP_METRIC_NAME,
    )
)


def get_builtin_metric_configs(
    extra_labels: frozenset[str],
    builtin_metrics_config: dict[str, dict[str, t.Any]],
) -> dict[str, MetricConfig]:
    """Return configuration for builtin metrics."""
    metric_configs = {
        metric_config.name: MetricConfig(
            metric_config.name,
            metric_config.description,
            metric_config.type,
            labels=set(metric_config.labels) | extra_labels,
            config=metric_config.config
            | builtin_metrics_config.get(metric_config.name, {}),
        )
        for metric_config in (
            _DB_ERRORS_METRIC_CONFIG,
            _QUERIES_METRIC_CONFIG,
            _QUERY_LATENCY_METRIC_CONFIG,
            _QUERY_TIMESTAMP_METRIC_CONFIG,
        )
    }
    metric_configs[QUERY_INTERVAL_METRIC_NAME] = _QUERY_INTERVAL_METRIC_CONFIG
    return metric_configs



"""Script entry point."""

from functools import partial
from pathlib import Path
import typing as t

from aiohttp.web import AppKey, Application
import click
from prometheus_aioexporter import (
    EXPORTER_APP_KEY,
    Arguments,
    InvalidMetricType,
    MetricConfig,
    PrometheusExporterScript,
)
from prometheus_client.metrics import Gauge

from . import __version__
from .config import (
    Config,
    ConfigError,
    load_config,
)
from .executor import QueryExecutor
from .metrics import QUERY_INTERVAL_METRIC_NAME

# The application key to track the QueryExecutor
QUERY_EXECUTOR_APP_KEY: AppKey[QueryExecutor] = AppKey("query-executor")


class QueryExporterScript(PrometheusExporterScript):
    """Periodically run database queries and export results to Prometheus."""

    name = "query-exporter"
    version = __version__
    description = __doc__
    default_port = 9560
    envvar_prefix = "QE"

    def command_line_parameters(self) -> list[click.Parameter]:
        return [
            click.Option(
                ["--check-only"],
                type=bool,
                help="only check configuration, don't run the exporter",
                is_flag=True,
                show_default=True,
                show_envvar=True,
            ),
            click.Option(
                ["--config"],
                type=click.Path(
                    exists=True,
                    dir_okay=False,
                    path_type=Path,
                ),
                help="configuration file",
                multiple=True,
                default=[Path("config.yaml")],
                show_default=True,
                show_envvar=True,
            ),
        ]

    def configure(self, args: Arguments) -> None:
        self.config = self._load_config(args.config)
        if args.check_only:
            self.logger.info("configuration valid")
            raise SystemExit(0)
        self.create_metrics(self.config.metrics.values())
        self._set_static_metrics()

    async def on_application_startup(
        self, application: Application
    ) -> None:  # pragma: nocover
        query_executor = QueryExecutor(self.config, self.registry, self.logger)
        application[EXPORTER_APP_KEY].set_metric_update_handler(
            partial(self._update_handler, query_executor)
        )
        application[QUERY_EXECUTOR_APP_KEY] = query_executor
        await query_executor.start()

    async def on_application_shutdown(
        self, application: Application
    ) -> None:  # pragma: nocover
        await application[QUERY_EXECUTOR_APP_KEY].stop()

    async def _update_handler(
        self, query_executor: QueryExecutor, metrics: list[MetricConfig]
    ) -> None:  # pragma: nocover
        """Run queries with no specified schedule on each request."""
        await query_executor.run_aperiodic_queries()
        query_executor.clear_expired_series()

    def _load_config(self, paths: list[Path]) -> Config:
        """Load the application configuration."""
        try:
            return load_config(paths, self.logger)
        except (InvalidMetricType, ConfigError) as error:
            self.logger.error("configuration invalid", error=str(error))
            if isinstance(error, ConfigError):
                for details in error.details:
                    self.logger.error("configuration invalid", **details)
            raise SystemExit(1)

    def _set_static_metrics(self) -> None:
        query_interval_metric = t.cast(
            Gauge, self.registry.get_metric(QUERY_INTERVAL_METRIC_NAME)
        )
        for query in self.config.queries.values():
            if query.interval:
                query_interval_metric.labels(query=query.name).set(
                    query.interval
                )


script = QueryExporterScript()



from asyncio import (
    Handle,
    Task,
    get_event_loop,
    iscoroutinefunction,
)
from collections.abc import (
    Callable,
    Iterator,
)
from functools import partial
import typing as t


class AlreadyRunning(Exception):
    """The TimedCall is already running."""

    def __init__(self) -> None:
        super().__init__("Timed call is already running")


class NotRunning(Exception):
    """The TimedCall is not running."""

    def __init__(self) -> None:
        super().__init__("Timed call is not running")


TimesIterator = Iterator[float | int]
FuncRetVal = t.TypeVar("FuncRetVal", bound=t.Any)


class TimedCall:
    """Call a function based on a timer.

    The class takes a function with optional arguments. Upon
    :meth:`start()`, the function is scheduled at specified times
    until :meth:`stop()` is called (or the time iterator is exausted).

    :param func: the function to call periodically.
    :param args: arguments to pass to the function.
    :param kwargs: keyword arguments to pass to the function.

    """

    def __init__(
        self, func: Callable[..., FuncRetVal], *args: t.Any, **kwargs: t.Any
    ) -> None:
        self._func = self._wrap_func(func, *args, **kwargs)
        self._loop = get_event_loop()
        self._handle: Handle | None = None
        self._task: Task[FuncRetVal] | None = None

    @property
    def running(self) -> bool:
        """Whether the PeriodicCall is currently running."""
        return self._handle is not None

    def start(self, times_iter: TimesIterator) -> None:
        """Start calling the function at specified times.

        :param times_iter: an iterable yielding times to execute the function
          at. If the iterator exhausts, the TimedCall is stopped.  Times must be
          compatible with :meth:`loop.time()`.

        """
        if self.running:
            raise AlreadyRunning()

        self._run(times_iter, do_call=False)

    async def stop(self) -> None:
        """Stop calling the function periodically."""
        if not self.running:
            raise NotRunning()

        if self._handle:
            self._handle.cancel()
            self._handle = None
        if self._task:
            await self._task
            self._task = None

    def _run(self, times_iter: TimesIterator, do_call: bool = True) -> None:
        if do_call:
            self._task = self._loop.create_task(self._func())
        self._schedule_next_run(times_iter)

    def _schedule_next_run(self, times_iter: TimesIterator) -> None:
        delay = self._get_run_delay(times_iter)
        if delay is None:
            self._handle = None
        else:
            self._handle = self._loop.call_later(delay, self._run, times_iter)

    def _get_run_delay(self, times_iter: TimesIterator) -> float | None:
        now = self._loop.time()
        next_time: float | int = -1
        while next_time < now:
            try:
                next_time = next(times_iter)
            except StopIteration:
                return None
        return float(next_time - now)

    def _wrap_func(
        self, func: Callable[..., FuncRetVal], *args: t.Any, **kwargs: t.Any
    ) -> Callable[[], t.Coroutine[t.Any, t.Any, FuncRetVal]]:
        if iscoroutinefunction(func):
            return t.cast(
                Callable[..., FuncRetVal], partial(func, *args, **kwargs)
            )
        else:

            async def f() -> FuncRetVal:
                return func(*args, **kwargs)

            return f


class PeriodicCall(TimedCall):
    """A TimedCall called at a fixed time intervals."""

    def start(self, interval: int | float, now: bool = True):  # type: ignore
        """Start calling the function periodically.

        :param interval: the time interval in seconds between calls.
        :param now: whether to make the first call immediately.

        """

        def times() -> Iterator[float]:
            time = self._loop.time()
            if not now:
                time += interval
            while True:
                yield time
                time += interval

        super().start(times())



"""Execute database queries and collect metrics."""

import asyncio
from collections import defaultdict
from collections.abc import Iterator, Mapping
from datetime import datetime
from itertools import chain
import time
import typing as t

from croniter import croniter
from dateutil.tz import gettz
from prometheus_aioexporter import (
    MetricConfig,
    MetricsRegistry,
)
from prometheus_client import Counter
from prometheus_client.metrics import MetricWrapperBase
import structlog

from .config import (
    DATABASE_LABEL,
    Config,
)
from .db import (
    Database,
    DatabaseConnectError,
    DatabaseError,
    MetricResult,
    Query,
    QueryExecution,
    QueryTimeoutExpired,
)
from .metrics import (
    DB_ERRORS_METRIC_NAME,
    QUERIES_METRIC_NAME,
    QUERY_LATENCY_METRIC_NAME,
    QUERY_TIMESTAMP_METRIC_NAME,
)
from .periodic_call import (
    PeriodicCall,
    TimedCall,
)


class InvalidMetricValue(Exception):
    """Raised when a query result is an invalid value for the metric."""


class MetricsLastSeen:
    """Track last seen times for metrics.

    It assumes labels are sorted by name in metrics.

    """

    def __init__(self, expirations: dict[str, int | None]):
        self._expirations = expirations
        self._last_seen: dict[str, dict[tuple[str, ...], float]] = defaultdict(
            dict
        )

    def update(
        self,
        name: str,
        labels: dict[str, str],
        timestamp: float,
    ) -> None:
        """Update last seen for a metric with a set of labels to given timestamp."""
        if not self._expirations.get(name):
            return

        # sort by label name
        label_values = tuple(value for _, value in sorted(labels.items()))
        self._last_seen[name][label_values] = timestamp

    def expire_series(
        self, timestamp: float
    ) -> dict[str, list[tuple[str, ...]]]:
        """Expire and return expired metric series at the given timestamp.

        Expired series are removed internally.

        Return a dict mapping metric names to a list of tuples of sorted label
        values for expired series.

        """
        expired = {}
        for name, metric_last_seen in self._last_seen.items():
            expiration = t.cast(int, self._expirations[name])
            expired_labels = [
                label_values
                for label_values, last_seen in metric_last_seen.items()
                if timestamp > last_seen + expiration
            ]
            if expired_labels:
                expired[name] = expired_labels

        # clear expired series from tracking
        for name, series_labels in expired.items():
            for label_values in series_labels:
                del self._last_seen[name][label_values]
                if not self._last_seen[name]:
                    del self._last_seen[name]
        return expired


class QueryExecutor:
    """Run database queries and collect metrics."""

    def __init__(
        self,
        config: Config,
        registry: MetricsRegistry,
        logger: structlog.stdlib.BoundLogger | None = None,
    ):
        self._config = config
        self._registry = registry
        self._logger = logger or structlog.get_logger()
        self._timed_query_executions: list[QueryExecution] = []
        self._aperiodic_query_executions: list[QueryExecution] = []
        # map query names to their TimedCalls
        self._timed_calls: dict[str, TimedCall] = {}
        # map query names to list of database names
        self._doomed_queries: dict[str, set[str]] = defaultdict(set)
        self._loop = asyncio.get_event_loop()
        self._last_seen = MetricsLastSeen(
            {
                name: metric.config.get("expiration")
                for name, metric in self._config.metrics.items()
            }
        )
        self._databases: dict[str, Database] = {
            name: Database(name, db_config, logger=self._logger)
            for name, db_config in self._config.databases.items()
        }

        for query_execution in chain(
            *(query.executions for query in self._config.queries.values())
        ):
            if query_execution.query.timed:
                self._timed_query_executions.append(query_execution)
            else:
                self._aperiodic_query_executions.append(query_execution)

    async def start(self) -> None:
        """Start timed queries execution."""
        for query_execution in self._timed_query_executions:
            query = query_execution.query
            call: TimedCall
            if query.interval:
                call = PeriodicCall(self._run_query, query_execution)
                call.start(query.interval, now=True)
            elif query.schedule is not None:
                call = TimedCall(self._run_query, query_execution)
                call.start(self._loop_times_iter(query.schedule))
            self._timed_calls[query_execution.name] = call

    async def stop(self) -> None:
        """Stop timed query execution."""
        coros = (call.stop() for call in self._timed_calls.values())
        await asyncio.gather(*coros, return_exceptions=True)
        self._timed_calls.clear()
        for db in self._databases.values():
            db.close()

    def clear_expired_series(self) -> None:
        """Clear metric series that have expired."""
        expired_series = self._last_seen.expire_series(self._loop.time())
        for name, label_values in expired_series.items():
            metric = self._registry.get_metric(name)
            for values in label_values:
                metric.remove(*values)

    async def run_aperiodic_queries(self) -> None:
        """Run queries on request."""
        coros = (
            self._execute_query(query_execution, dbname)
            for query_execution in self._aperiodic_query_executions
            for dbname in query_execution.query.databases
        )
        await asyncio.gather(*coros, return_exceptions=True)

    def _loop_times_iter(self, schedule: str) -> Iterator[float | int]:
        """Wrap a croniter iterator to sync time with the loop clock."""
        cron_iter = croniter(schedule, datetime.now(gettz()))
        while True:
            cc = next(cron_iter)
            t = time.time()
            delta = cc - t
            yield self._loop.time() + delta

    def _run_query(self, query_execution: QueryExecution) -> None:
        """Periodic task to run a query."""
        for dbname in query_execution.query.databases:
            self._loop.create_task(
                self._execute_query(query_execution, dbname)
            )

    async def _execute_query(
        self, query_execution: QueryExecution, dbname: str
    ) -> None:
        """'Execute a Query on a Database."""
        if await self._remove_if_dooomed(query_execution, dbname):
            return

        db = self._databases[dbname]
        query = query_execution.query
        try:
            metric_results = await db.execute(query_execution)
            if metric_results.latency:
                self._update_query_latency_metric(
                    db, query, metric_results.latency
                )
            if metric_results.timestamp:
                self._update_query_timestamp_metric(
                    db, query, metric_results.timestamp
                )
            self._update_metrics_from_results(
                db, query_execution.name, metric_results.results
            )
            self._increment_queries_count(db, query, "success")
        except InvalidMetricValue:
            self._increment_queries_count(db, query, "invalid-value")
        except DatabaseConnectError:
            self._increment_db_error_count(db)
        except QueryTimeoutExpired:
            self._increment_queries_count(db, query, "timeout")
        except DatabaseError as error:
            self._increment_queries_count(db, query, "error")
            if error.fatal:
                self._logger.debug(
                    "removing failed query",
                    query=query_execution.name,
                    database=dbname,
                )
                self._doomed_queries[query_execution.name].add(dbname)

    async def _remove_if_dooomed(
        self, query_execution: QueryExecution, dbname: str
    ) -> bool:
        """Remove a query execution if it will never work.

        Return whether the query has been removed for the database.

        """
        if dbname not in self._doomed_queries[query_execution.name]:
            return False

        query = query_execution.query

        if set(query.databases) == self._doomed_queries[query_execution.name]:
            # the query has failed on all databases
            if query.timed:
                self._timed_query_executions.remove(query_execution)
                call = self._timed_calls.pop(query_execution.name, None)
                if call is not None:
                    await call.stop()
            else:
                self._aperiodic_query_executions.remove(query_execution)
        return True

    def _update_metrics_from_results(
        self,
        database: Database,
        query_execution_name: str,
        results: list[MetricResult],
    ) -> None:
        has_invalid = False
        for result in results:
            try:
                self._update_metric(
                    database, result.metric, result.value, labels=result.labels
                )
            except ValueError as e:
                self._logger.debug(
                    "invalid metric result",
                    error=str(e),
                    query=query_execution_name,
                    database=database.name,
                )
                has_invalid = True

        # only raise the error after processing all values
        if has_invalid:
            raise InvalidMetricValue()

    def _update_metric(
        self,
        database: Database,
        name: str,
        value: t.Any,
        labels: Mapping[str, str] | None = None,
    ) -> None:
        """Update value for a metric."""
        if value is None:
            # count queries might return NULL, treat it as zero
            value = 0.0
        metric_config = self._config.metrics[name]
        all_labels = {DATABASE_LABEL: database.name}
        all_labels.update(database.config.labels)
        if labels:
            all_labels.update(labels)
        method = self._get_metric_method(metric_config)
        self._logger.debug(
            "updating metric",
            metric=name,
            method=method,
            value=value,
            labels=all_labels,
        )
        metric = self._registry.get_metric(name, labels=all_labels)
        self._update_metric_value(metric, method, value)
        self._last_seen.update(name, all_labels, self._loop.time())

    def _get_metric_method(self, metric: MetricConfig) -> str:
        if metric.type == "counter" and not metric.config.get(
            "increment", False
        ):
            method = "set"
        else:
            method = {
                "counter": "inc",
                "gauge": "set",
                "histogram": "observe",
                "summary": "observe",
                "enum": "state",
            }[metric.type]
        return method

    def _update_metric_value(
        self, metric: MetricWrapperBase, method: str, value: t.Any
    ) -> None:
        if metric._type == "counter" and method == "set":
            # counters can only be incremented, directly set the underlying value
            t.cast(Counter, metric)._value.set(value)
        else:
            getattr(metric, method)(value)

    def _increment_queries_count(
        self, database: Database, query: Query, status: str
    ) -> None:
        """Increment count of queries in a status for a database."""
        self._update_metric(
            database,
            QUERIES_METRIC_NAME,
            1,
            labels={"query": query.name, "status": status},
        )

    def _increment_db_error_count(self, database: Database) -> None:
        """Increment number of errors for a database."""
        self._update_metric(database, DB_ERRORS_METRIC_NAME, 1)

    def _update_query_latency_metric(
        self, database: Database, query: Query, latency: float
    ) -> None:
        """Update latency metric for a query on a database."""
        self._update_metric(
            database,
            QUERY_LATENCY_METRIC_NAME,
            latency,
            labels={"query": query.name},
        )

    def _update_query_timestamp_metric(
        self, database: Database, query: Query, timestamp: float
    ) -> None:
        """Update timestamp metric for a query on a database."""
        self._update_metric(
            database,
            QUERY_TIMESTAMP_METRIC_NAME,
            timestamp,
            labels={"query": query.name},
        )



import os
from pathlib import Path
import typing as t

import yaml


def load_yaml(path: Path) -> t.Any:
    """Load a YAML document from a file."""

    with path.open() as fd:
        return yaml.load(fd, _config_loader(path.parent))


class _ConfigLoader(yaml.SafeLoader):
    """YAML loader supporting tags."""

    base_path: t.ClassVar[Path]


def _config_loader(path: Path) -> type[_ConfigLoader]:
    class ConfigLoaderWithPath(_ConfigLoader):
        base_path = path

    return ConfigLoaderWithPath


def _tag_env(loader: _ConfigLoader, node: yaml.nodes.ScalarNode) -> t.Any:
    env = loader.construct_scalar(node)
    value = os.getenv(env)
    if value is None:
        raise yaml.scanner.ScannerError(
            "while processing 'env' tag",
            None,
            f"variable {env} undefined",
            loader.get_mark(),  # type: ignore
        )
    return yaml.safe_load(value)


def _tag_file(loader: _ConfigLoader, node: yaml.nodes.ScalarNode) -> str:
    path = loader.base_path / loader.construct_scalar(node)
    if not path.is_file():
        raise yaml.scanner.ScannerError(
            "while processing 'file' tag",
            None,
            f"file {path} not found",
            loader.get_mark(),  # type: ignore
        )
    return path.read_text().strip()


def _tag_include(loader: _ConfigLoader, node: yaml.nodes.ScalarNode) -> t.Any:
    path = loader.base_path / loader.construct_scalar(node)
    if not path.is_file():
        raise yaml.scanner.ScannerError(
            "while processing 'include' tag",
            None,
            f"file {path} not found",
            loader.get_mark(),  # type: ignore
        )
    with path.open() as fd:
        return yaml.load(fd, _config_loader(path.parent))


_ConfigLoader.add_constructor("!env", _tag_env)
_ConfigLoader.add_constructor("!file", _tag_file)
_ConfigLoader.add_constructor("!include", _tag_include)



"""Database wrapper."""

import asyncio
from collections.abc import (
    Iterable,
    Iterator,
    Sequence,
)
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from dataclasses import (
    InitVar,
    dataclass,
    field,
)
from itertools import chain
from time import (
    perf_counter,
    time,
)
import typing as t

from croniter import croniter
from sqlalchemy import (
    create_engine,
    event,
    make_url,
    text,
)
from sqlalchemy.engine import (
    CursorResult,
    Engine,
)
from sqlalchemy.engine.interfaces import DBAPIConnection, DBAPICursor
from sqlalchemy.exc import (
    ArgumentError,
    NoSuchModuleError,
)
from sqlalchemy.pool import ConnectionPoolEntry, QueuePool
import structlog

from . import schema

# Timeout for a query
QueryTimeout = int | float


class DatabaseError(Exception):
    """A databease error.

    if `fatal` is True, it means the Query will never succeed.
    """

    def __init__(self, message: str, fatal: bool = False) -> None:
        super().__init__(message)
        self.fatal = fatal


class DatabaseConnectError(DatabaseError):
    """Database connection error."""


class DatabaseQueryError(DatabaseError):
    """Database query error."""


class QueryTimeoutExpired(Exception):
    """Query execution timeout expired."""


class InvalidResultCount(Exception):
    """Number of results from a query don't match metrics count."""

    def __init__(self, expected: int, got: int) -> None:
        super().__init__(
            f"Wrong result count from query: expected {expected}, got {got}"
        )


class InvalidResultColumnNames(Exception):
    """Invalid column names in query results."""

    def __init__(self, expected: list[str], got: list[str]) -> None:
        super().__init__(
            "Wrong column names from query: "
            f"expected {self._names(expected)}, got {self._names(got)}"
        )

    def _names(self, names: list[str]) -> str:
        names_list = ", ".join(names)
        return f"({names_list})"


class InvalidQueryParameters(Exception):
    """Query parameter names don't match those in query SQL."""

    def __init__(self, query_name: str) -> None:
        super().__init__(
            f'Parameters for query "{query_name}" don\'t match those from SQL'
        )


class InvalidQuerySchedule(Exception):
    """Query schedule is wrong or both schedule and interval specified."""

    def __init__(self, query_name: str, message: str) -> None:
        super().__init__(
            f'Invalid schedule for query "{query_name}": {message}'
        )


# Database errors that mean the query won't ever succeed.  Not all possible
# fatal errors are tracked here, because some DBAPI errors can happen in
# circumstances which can be fatal or not.  Since there doesn't seem to be a
# reliable way to know, there might be cases when a query will never succeed
# but will end up being retried.
FATAL_ERRORS = (InvalidResultCount, InvalidResultColumnNames)


def create_db_engine(config: schema.Database) -> Engine:
    """Create the database engine, validating the configuration."""
    try:
        url = make_url(str(config.dsn))
        connect_args = {}
        if url.get_backend_name() == "sqlite":
            # disable check as the logic ensures connections aren't used by
            # multiple threads at once
            connect_args["check_same_thread"] = False

        return create_engine(
            url,
            poolclass=QueuePool,
            pool_size=config.connection_pool.size,
            max_overflow=config.connection_pool.max_overflow,
            pool_pre_ping=True,
            pool_recycle=3600,
            connect_args=connect_args,
        )
    except ImportError as error:
        raise DatabaseError(f'Module "{error.name}" not found')
    except (ArgumentError, ValueError, NoSuchModuleError):
        raise DatabaseError(f'Invalid database DSN: "{config.dsn}"')


class QueryMetric(t.NamedTuple):
    """Metric details for a Query."""

    name: str
    labels: Iterable[str]


class QueryResults(t.NamedTuple):
    """Results of a database query."""

    keys: list[str]
    rows: Sequence[Sequence[t.Any]]
    timestamp: float | None = None
    latency: float | None = None

    @classmethod
    def from_result(cls, result: CursorResult[t.Any]) -> t.Self:
        """Return a QueryResults from results for a query."""
        timestamp = time()
        keys: list[str] = []
        rows: Sequence[Sequence[t.Any]] = []
        if result.returns_rows:
            keys, rows = list(result.keys()), result.all()
        latency = result.connection.info.get("query_latency", None)
        return cls(keys, rows, timestamp=timestamp, latency=latency)


class MetricResult(t.NamedTuple):
    """A result for a metric from a query."""

    metric: str
    value: t.Any
    labels: dict[str, str]


class MetricResults(t.NamedTuple):
    """Collection of metric results for a query."""

    results: list[MetricResult]
    timestamp: float | None = None
    latency: float | None = None


@dataclass
class Query:
    """Query definition and configuration."""

    name: str
    databases: list[str]
    metrics: list[QueryMetric]
    sql: str
    timeout: QueryTimeout | None = None
    interval: int | None = None
    schedule: str | None = None

    parameter_sets: InitVar[list[dict[str, t.Any]] | None] = None
    executions: list["QueryExecution"] = field(init=False, compare=False)

    def __post_init__(
        self, parameter_sets: list[dict[str, t.Any]] | None
    ) -> None:
        self._check_schedule()
        if not parameter_sets:
            self.executions = [QueryExecution(self.name, self)]
        else:
            self.executions = [
                QueryExecution(f"{self.name}[params{index}]", self, parameters)
                for index, parameters in enumerate(parameter_sets, 1)
            ]

    @property
    def timed(self) -> bool:
        """Whether the query is run periodically via interval or schedule."""
        return bool(self.interval or self.schedule)

    def labels(self) -> frozenset[str]:
        """Resturn all labels for metrics in the query."""
        return frozenset(chain(*(metric.labels for metric in self.metrics)))

    def results(self, query_results: QueryResults) -> MetricResults:
        """Return MetricResults from a query."""
        if not query_results.rows:
            return MetricResults([])

        result_keys = sorted(query_results.keys)
        labels = self.labels()
        metrics = [metric.name for metric in self.metrics]
        expected_keys = sorted(set(metrics) | labels)
        if len(expected_keys) != len(result_keys):
            raise InvalidResultCount(len(expected_keys), len(result_keys))
        if result_keys != expected_keys:
            raise InvalidResultColumnNames(expected_keys, result_keys)
        results = []
        for row in query_results.rows:
            values = dict(zip(query_results.keys, row))
            for metric in self.metrics:
                metric_result = MetricResult(
                    metric.name,
                    values[metric.name],
                    {label: values[label] for label in metric.labels},
                )
                results.append(metric_result)

        return MetricResults(
            results,
            timestamp=query_results.timestamp,
            latency=query_results.latency,
        )

    def _check_schedule(self) -> None:
        if self.interval and self.schedule:
            raise InvalidQuerySchedule(
                self.name, "both interval and schedule specified"
            )
        if self.schedule and not croniter.is_valid(self.schedule):
            raise InvalidQuerySchedule(self.name, "invalid schedule format")


@dataclass(frozen=True)
class QueryExecution:
    """A single execution configuration for a query, with parameters."""

    name: str
    query: Query
    parameters: dict[str, t.Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._check_query_parameters()

    def _check_query_parameters(self) -> None:
        expr = text(self.query.sql)
        query_params = set(expr.compile().params)
        if set(self.parameters) != query_params:
            raise InvalidQueryParameters(self.name)


class Database:
    """A database to perform Queries."""

    def __init__(
        self,
        name: str,
        config: schema.Database,
        logger: structlog.stdlib.BoundLogger | None = None,
    ) -> None:
        self.name = name
        self.config = config
        if logger is None:
            logger = structlog.get_logger()
        self.logger = logger.bind(database=self.name)

        self._engine = self._setup_engine()

        pool_config = self.config.connection_pool
        max_workers = pool_config.size + pool_config.max_overflow
        self._executor = ThreadPoolExecutor(
            max_workers=max_workers,
            thread_name_prefix=f"[Database-{self.name}]",
        )

    def close(self) -> None:
        self._engine.dispose()
        self._executor.shutdown(wait=True, cancel_futures=True)

    async def execute(self, query_execution: QueryExecution) -> MetricResults:
        """Execute a query."""
        self.logger.debug("run query", query=query_execution.name)
        query = query_execution.query
        try:
            query_results = await self.execute_sql(
                query.sql,
                parameters=query_execution.parameters,
                timeout=query.timeout,
            )
            return query.results(query_results)
        except DatabaseError:
            raise
        except TimeoutError:
            self.logger.warning("query timeout", query=query_execution.name)
            raise QueryTimeoutExpired()
        except Exception as error:
            raise self._query_db_error(
                query_execution.name,
                error,
                fatal=isinstance(error, FATAL_ERRORS),
            )

    async def execute_sql(
        self,
        sql: str,
        parameters: dict[str, t.Any] | None = None,
        timeout: QueryTimeout | None = None,
    ) -> QueryResults:
        """Execute a raw SQL query."""
        if parameters is None:
            parameters = {}

        loop = asyncio.get_event_loop()
        return await asyncio.wait_for(
            loop.run_in_executor(
                self._executor, self._execute_sync, sql, parameters
            ),
            timeout=timeout,
        )

    def _setup_engine(self) -> Engine:
        engine = create_db_engine(self.config)

        @event.listens_for(engine, "connect")
        def on_connect(
            dbapi_conn: DBAPIConnection, conn: ConnectionPoolEntry
        ) -> None:
            if self.config.connect_sql:
                try:
                    with self._dbapi_transaction(dbapi_conn) as cursor:
                        for sql in self.config.connect_sql:
                            cursor.execute(sql)
                except Exception as error:
                    raise self._db_error(
                        f"failed executing connect SQL: {error}",
                        exc_class=DatabaseQueryError,
                    )
            self.logger.debug("connected")

        @event.listens_for(engine, "close")
        def on_close(
            dbapi_conn: DBAPIConnection, conn: ConnectionPoolEntry
        ) -> None:
            conn.info.clear()
            self.logger.debug("disconnected")

        @event.listens_for(engine, "before_cursor_execute")
        def before_cursor_execute(
            conn: ConnectionPoolEntry,
            cursor: t.Any,
            statement: str,
            parameters: t.Any,
            context: t.Any,
            executemany: bool,
        ) -> None:
            conn.info["query_start_time"] = perf_counter()

        @event.listens_for(engine, "after_cursor_execute")
        def after_cursor_execute(
            conn: ConnectionPoolEntry,
            cursor: t.Any,
            statement: str,
            parameters: t.Any,
            context: t.Any,
            executemany: bool,
        ) -> None:
            conn.info["query_latency"] = perf_counter() - conn.info.pop(
                "query_start_time"
            )

        return engine

    def _execute_sync(
        self,
        sql: str,
        parameters: dict[str, t.Any],
    ) -> QueryResults:
        try:
            with self._engine.begin() as conn:
                try:
                    result = conn.execute(text(sql), parameters)
                    return QueryResults.from_result(result)
                except Exception as error:
                    raise self._db_error(error, exc_class=DatabaseQueryError)
        except DatabaseQueryError:
            raise
        except Exception as error:
            raise self._db_error(error, exc_class=DatabaseConnectError)

    def _query_db_error(
        self,
        query_name: str,
        error: str | Exception,
        fatal: bool = False,
    ) -> DatabaseError:
        """Create and log a DatabaseError for a failed query."""
        message = self._error_message(error)
        self.logger.exception(
            "query failed", query=query_name, error=message, exception=error
        )
        return DatabaseQueryError(message, fatal=fatal)

    def _db_error(
        self,
        error: str | Exception,
        exc_class: type[DatabaseError] = DatabaseError,
        fatal: bool = False,
    ) -> DatabaseError:
        """Create and log a DatabaseError."""
        message = self._error_message(error)
        self.logger.exception("database error", exception=error)
        return exc_class(message, fatal=fatal)

    def _error_message(self, error: str | Exception) -> str:
        """Return a message from an error."""
        message = str(error).strip()
        if not message and isinstance(error, Exception):
            message = error.__class__.__name__
        return message

    @contextmanager
    def _dbapi_transaction(
        self, dbapi_conn: DBAPIConnection
    ) -> Iterator[DBAPICursor]:
        cursor = dbapi_conn.cursor()
        try:
            yield cursor
            dbapi_conn.commit()
        except Exception:
            dbapi_conn.rollback()
            raise
        finally:
            cursor.close()
